#### Tutorial on Uncertainty Quantification in Machine Learning
####
####  Boston Consulting Group,  April 6/7, 2022

## The data throughout were downloaded from 
##
## https://www.kaggle.com/datasets/ruiqurm/lianjia
##
 
# Here we make use of the following libraries

library('ranger')
library('randomForest')
library('iml')
library('shapr')
library('mgcv')
library('rpart')
library('rpart.plot')
library('glmnet')

# Loading in the Beijing Housing data

data = read.csv('BeijingRealEstate2.csv')

# Make sure categorical features are categorical
faccols = c(11,13,14,16:19)
for(i in 1:length(faccols)){ data[,faccols[i]] = as.factor(data[,faccols[i]]) }

# But also that categorical columns are categorical
data$constructionTime = as.numeric(data$constructionTime)
data$livingRoom = as.numeric(data$livingRoom)
data$drawingRoom = as.numeric(data$drawingRoom)
data$kitchen = as.numeric(data$kitchen)
data$bathRoom = as.numeric(data$bathRoom)


# Remove missing values

data = data[apply(is.na(data),1,sum)==0,]

# Re-order columns

p = ncol(data)
data0 = data  #[,c(5,(1:p)[-5])]

# Select a small set of data at random

set.seed('05032022')
n = 4500
index = sample(nrow(data0),n,replace=FALSE)
data = data0[index,]

# And split this into test and train

tindex = sample(n,floor(2*n/3),replace=FALSE)
test = data[-tindex,]
train = data[tindex,]


# Later on it will be helpful to have all-numeric data. Since we won't care too
# much about the specifics of the features we'll just turn them into 
# numeric values directly. 

ntrain = train
ntest = test
for(j in 1:p){ ntrain[,j] = as.numeric(ntrain[,j]); ntest[,j] = as.numeric(ntest[,j]) }


###########################################
# Let's start off with a basic random forest. 

mod1 = randomForest(log(totalPrice)~.,data=train,importance=TRUE)

# Here is the iml library applied to it to get importances

mod1.interp = Predictor$new(mod1,data=test[,-1],y=test$totalPrice)
mod1.imp = FeatureImp$new(mod1.interp,loss='mse')
plot(mod1.imp)

# and feature effects for 'square'

plot(FeatureEffect$new(mod1.interp,5,method='pdp+ice'))
plot(FeatureEffect$new(mod1.interp,5,method='ale'))


# But so we know what we're doing explicitly,let's look at
# change in OOB MSE when we re-train models with each of the 
# features dropped

DropVarImp = rep(NA,p-1)
names(DropVarImp) = names(data)[2:p]

for(j in 1:(p-1)){
 tempmod = randomForest(log(totalPrice)~.,data=train[,-(j+1)])

 DropVarImp[j] = tempmod$mse[500] - mod1$mse[500]
}


par(mar=c(10,4,1,1))
barplot(DropVarImp,las=2,col='darkblue')

### Shapley Values

# Calculating Shapley values for change in MSE. Here we'll examine
# 100 orderings of the data. 

# This code calculates Shapley values for a number of differences
#  - OOB MSE measures
#  - predictions at two different test points
#  - predictions from a model trained by ranger at two test points
#
# We use both randomForest and ranger because they interract with different
# explanation packages

nperm  = 100
OrdMat = matrix(NA,nperm,p-1)  # Which ordering to use

DVMat = matrix(NA,nperm,p-1)  # This records MSE
DPredMat = matrix(NA,nperm,p-1)  # Predictions at an individual data point
DPredMat2 = matrix(NA,nperm,p-1) # And at a second
rDPredMat = matrix(NA,nperm,p-1) # Predictions from a model trained by ranger
rDPredMat2 = matrix(NA,nperm,p-1) # and a second

# The arrays above store successive differences in the order
# given by OrdMat. Below we re-order to correspond to the 
# order of features in the data. 

DropVarDiff = matrix(NA,nperm,p-1)
DropPredDiff = matrix(NA,nperm,p-1)
DropPredDiff2 = matrix(NA,nperm,p-1)
rDropPredDiff = matrix(NA,nperm,p-1)
rDropPredDiff2 = matrix(NA,nperm,p-1)



# Note that this takes a while, and we might want to just load a 
# set of results. 

 for(k in 1:nperm){
	 OrdMat[k,] = sample(2:p,replace=FALSE)

	 # Record MSEs and predictions for models as you include each 
	 # feature in turn. 
	 
	 # We use ranger for predictions because the shapr package supports ranger
	 # predictions natively
	 for(j in 1:(p-1)){
		 # Train randomForest and ranger models
		 tempmod = randomForest(log(totalPrice)~.,data=train[,c(1,OrdMat[k,1:j])])
		 rtempmod = ranger(log(totalPrice)~.,data=ntrain[,c(1,OrdMat[k,1:j])])
		 
		 # Record MSE and Predictions
		 DVMat[k,j] = tempmod$mse[500]
		 DPredMat[k,j] = predict(tempmod,test[1,])
		 DPredMat[k,j] = predict(tempmod,test[2,])
		 rDPredMat[k,j] = predict(rtempmod,ntest[1,])$predictions
		 rDPredMat2[k,j] = predict(rtempmod,ntest[2,])$predictions
	 }
	 
	 # Take differences; note that we expect MSE's to _decrease_ along the way.
	 # We also need to order them according to the feature rather than the sequence
	 # of inclusion
	 
	 DVMat[k,] = c( DVMat[k,1] - mod1$mse[1], -diff(DVMat[k,]) )
	 DropVarDiff[k,OrdMat[k,]-1] = DVMat[k,]
	 
	 DPredMat[k,] = c( DPredMat[k,1] - mean(log(test$totalPrice)), diff(DPredMat[k,]) )
	 DropPredDiff[k,OrdMat[k,]-1] = DPredMat[k,]
	 
	 DPredMat2[k,] = c( DPredMat2[k,1] - mean(log(test$totalPrice)), diff(DPredMat2[k,]) )
	 DropPredDiff2[k,OrdMat[k,]-1] = DPredMat2[k,]
	 
	 rDPredMat[k,] = c( rDPredMat[k,1] - mean(log(test$totalPrice)), diff(rDPredMat[k,]) )
	 rDropPredDiff[k,OrdMat[k,]-1] = rDPredMat[k,]
	 
	 rDPredMat2[k,] = c( rDPredMat2[k,1] - mean(log(test$totalPrice)), diff(rDPredMat2[k,]) )
	 rDropPredDiff2[k,OrdMat[k,]-1] = rDPredMat2[k,]
	 print(k)
 }


# Now let's look at mean and standard deviations of the changes. 
# Formally, the mean is a Monte Carlo estimate of the Shapley value

ShapVarImp = apply(DropVarDiff,2,mean)
ShapSD = apply(DropVarDiff,2,sd)

# This is the standard Shapley plot for global variable importances
names(ShapVarImp) = names(DropVarImp)
par(mar=c(10,4,1,1))
barplot(ShapVarImp,col='darkblue',ylab='',las=2)

# I think it helps to plot the differences for each 
# ordering of the features

matplot(t(DropVarDiff),type='l',col='grey',xaxt='n',ylab='')
axis(1,1:length(DropVarImp),names(DropVarImp),las=2,cex.axis=1)
lines(ShapVarImp,col='darkblue',lwd=5)

# Features 6 and 7 seem to have importance values that are particularly #
# sensitive to the ordering in which they come.  Let's look at which
# permutations made them look most important. 

w.perm6 = (1:nperm)[OrdMat[,p-2]==6][which.max( DVMat[OrdMat[,p-2]==6,p-2])]
OrdMat[w.perm6,]


w.perm7 = (1:nperm)[OrdMat[,p-2]==7][which.max( DVMat[OrdMat[,p-2]==7,p-2])]
OrdMat[w.perm7,]


# Let's do Drop-and-retrain feature importance with each of 
# Feature 6 and 7 removed to see how importance changes in their absense 

DropVarImp6 = rep(NA,p-1)

temptrain = train[,-6]

mod6 = randomForest(log(totalPrice)~.,data=temptrain)
for(j in 1:(p-2)){
 tempmod = randomForest(log(totalPrice)~.,data=temptrain[,-(j+1)])

 DropVarImp6[j] = tempmod$mse[500] - mod6$mse[500]
}

# Now dropping feature 7

DropVarImp7 = rep(NA,p-1)

temptrain = train[,-7]

mod7 = randomForest(log(totalPrice)~.,data=temptrain)
for(j in 1:(p-2)){
 tempmod = randomForest(log(totalPrice)~.,data=temptrain[,-(j+1)])

 DropVarImp7[j] = tempmod$mse[500] - mod7$mse[500]
}


# If we plot these up we see how feature importance changes with 
# the reference set. 

barplot(rbind(DropVarImp,
	c(DropVarImp6[1:4],0,DropVarImp6[5:(p-2)]),
	c(DropVarImp7[1:5],0,DropVarImp6[6:(p-2)])),beside=TRUE,col=c('darkblue','red','darkgreen'),las=2)

plot(DropVarImp)
lines(c(DropVarImp6[1:4],NA,DropVarImp6[5:(p-2)]),col=2)
lines(c(DropVarImp7[1:5],NA,DropVarImp6[6:(p-2)]),col=4)
legend('topright',c('all','-square','-livingRoom'),lty=1,lwd=4,cex=2,col=c('darkblue','red','darkgreen'))

par(mar=c(5,5,1,1))
plot(train$square,train$livingRoom,pch='*',col='darkblue',cex=2,xlab='square',ylab='livingRoom',cex.lab=2,cex.axis=1.5)


# out of curiosity, how does Dropped feature importance compare to the permutation
# importance you get from randomForest? We'll need to re-scale because their amplitudes
# don't match 

par(mar=c(10,4,1,1))
barplot(t(cbind(DropVarImp/sum(DropVarImp),mod1$importance[,1]/sum(mod1$importance[,1]))),col=c('darkblue','red'),beside=TRUE,las=2)

### Ground-truthing feature importances. Do you get something that looks like what
### you'd get if you knew the truth?

# We'll turn that feature importance into coefficients in a linear model

coef = sqrt(mod1$importance[,1]/apply(ntrain[,-1],2,var))

# Here is where we will use the all-numeric matrix, in order to create a linear model.
# Note that we are creating a fake value of totalPrice:

ntrain$totalPrice = as.matrix(ntrain[,-1])%*%coef

# and we'll train a randomForest on that linear model

tempmod = randomForest(totalPrice~.,data=ntrain,importance=TRUE)

## We can run our own permutation improtance code pretty easily, just 
# permute one column in turn and see how the predictions change. 


PermImpLin = rep(NA,p-1)  # What happens in the true linear model
PermImpRF = rep(NA,p-1)   # What happens in the random forest

for(j in 1:p-1){
	# Training set with one column permuted
	tntrain = ntrain
	tntrain[,j+1] = sample(tntrain[,j+1], replace=FALSE)
	
	PermImpLin[j] = mean( (ntrain$totalPrice - as.matrix(tntrain[,-1])%*%coef)^2 )
	PermImpRF[j] = mean( (ntrain$totalPrice - predict(tempmod,tntrain))^2 )
}

# If we look at importances, Random Forests has over-emphasized a couple of features
names(PermImpLin) = names(DropVarImp)
names(PermImpRF) = names(DropVarImp)
par(mar=c(10,4,1,1))
barplot( rbind(PermImpLin,PermImpRF),ylab='',beside=TRUE,las=2,col=c('darkblue','red'))


# Looking at the correlation between square and bathroom, what has happened is that
# permutation methods make a lot of predictions where we have no data, and randomForests
# extrapolates in a way that emphasizes them. 

par(mar=c(5,5,1,1))
plot(train$square,train$bathRoom,xlab='square',ylab='bathRoom',cex.lab=2,cex.axis=2,pch='*',cex=2,col='darkblue')
points( rep(train$square,3000),rep(train$bathRoom,1,each=3000),pch='o',col='red')
points(train$square,train$bathRoom,pch='*',cex=2,col='darkblue')

# And we can also look at partial dependence and ICE plots for what should be a linear
# response. Here we see some attenuation at the edges, also from the same effects. 

mod1.interp = Predictor$new(tempmod,data=ntrain[,-1],y=ntrain[,1])
plot(FeatureEffect$new(mod1.interp,5,method='pdp+ice'))
plot(FeatureEffect$new(mod1.interp,5,method='ale'))

##### Model Distillation 
#
# Here we will build a GAM to predict the output of mod1. 

temptrain = train
temptrain$totalPrice = predict(mod1,train)

gammod = gam(totalPrice~s(Lng)+s(Lat)+s(DOM)+s(followers)+s(square)+
                  livingRoom+drawingRoom+kitchen+bathRoom+buildingType+s(constructionTime)+
				  renovationCondition+buildingStructure+ladderRatio+elevator+fiveYearsProperty+subway,data=temptrain)
par(mfrow=c(2,3))
plot(gammod,cex.lab=1.5,cex.axis=1.5)


# Now if we try to build a GAM to predict a linear response, we get something that
# looks pretty good. 

gammod2 = gam(totalPrice~s(Lng)+s(Lat)+s(DOM)+s(followers)+s(square)+
                  livingRoom+drawingRoom+kitchen+bathRoom+buildingType+s(constructionTime)+
				  renovationCondition+buildingStructure+ladderRatio+elevator+fiveYearsProperty+subway,data=ntrain)

par(mfrow=c(2,3))
plot(gammod2,cex.lab=1.5,cex.axis=1.5)
	
	
#### What if we distill into two different CART trees?

# Here we'll make use of the huge amount of unused data to create
# two sets of 20,000 data points

psudat1 = data0[-index,][1:20000,]
psudat2 = data0[-index,][20000+(1:20000),]

# and associate the predictions with mod1 from them. 

psudat1$totalPrice = predict(mod1,psudat1)
psudat2$totalPrice = predict(mod1,psudat2)

# Lets build CART models

cartmod1 = rpart(totalPrice~.,data=psudat1,cp=0,maxdepth=3)
cartmod2 = rpart(totalPrice~.,data=psudat2,cp=0,maxdepth=3)

# Plotting these reveals differences at the third level, that get
# worse further down. 

rpart.plot(cartmod1)
rpart.plot(cartmod2)



### LIME as distillation. 

# LIME fits a local linear model with a LASSO penalty
# We'll do this by looking at 1000 points nearby a point
# of interest and using glmnet for the LASSO. 

# Let's find a test point

t1 = test[1,]

# and look for the nearest 100 data points in psudat1 to it


D1 = dist(rbind(t1[-1],psudat1[,-1]),method='euclidean')
loc1 = psudat1[sort(D1[2:20001],index.return=TRUE)$ix[1:1000],]

# We'll associate predictions from the random forest as the response

loc1$totalPrice = predict(mod1,loc1)
loc1mat = model.matrix(totalPrice~.,data=loc1)

# Fit a LASSO to these
lime1 = glmnet(loc1mat,loc1$totalPrice)

# And find the point at which we have 5 nonzero coefficients
pos1 = max((1:length(lime1$df))[lime1$df < 6])

# Let's look at these:
beta1 = lime1$beta[,pos1]
beta1 = beta1[beta1 !=0]

# We'll do the same thing for psudat2

D2 = dist(rbind(t1[-1],psudat2[,-1]),method='euclidean')
loc2 = psudat1[sort(D2[2:20001],index.return=TRUE)$ix[1:1000],]
loc2$totalPrice = predict(mod1,loc2)

loc2mat = model.matrix(totalPrice~.,data=loc2)
lime2 = glmnet(loc2mat,loc2$totalPrice)

pos2 = max((1:length(lime2$df))[lime2$df < 6])
beta2 = lime2$beta[,pos2]
beta2 = beta2[beta2 !=0]




### SHAP values for individual predictions

# Ranger works with the shapr package, so let's go back to our original
# data and produce a random forest model with it.

rmod = ranger(y=log(train$totalPrice),x=ntrain[,-1],importance='permutation')
ntest = data.frame(ntest)

# Here we get SHAP explanations out of shapr, these use a kernel conditional density
# to average predictions over the conditional value of each feature. 

explainer = shapr(ntest[,-1],rmod,n_combinations=100)
y0 = mean(log(train$totalPrice))
shap1 = shapr::explain(ntest[1:2,-1],explainer,approach="empirical",prediction_zero=y0)

# Here are the SHAP values with independence assumptions, so we average over the marginal
# distribution of each feature

shap2 = shapr::explain(ntest[1:2,-1],explainer,approach="empirical",prediction_zero=y0,type='independence')


# We also get re-learned SHAP values out of rDropPredDiff from up above. 

# For point 1
par(mar=c(10,5,1,1))
matplot(cbind(t(shap1$dt[1,-1]),t(shap2$dt[1,-1]),-apply(rDropPredDiff,2,mean)),pch='*',
	col=c('blue','red','black'),type='b',lwd=2,lty=1,xaxt='n',ylab='SHAP',
	cex.lab=1.5,cex.axis=1.5,cex=2)
axis(1,1:length(DropVarImp),names(DropVarImp),las=2,cex.axis=1)
abline(h=0)
legend('topright',c('Kernel','Independence','Drop'),lwd=2,col=c('blue','red','black'),lty=1,cex=1.5)

# And for point 2
par(mar=c(10,5,1,1))
matplot(cbind(t(shap1$dt[2,-1]),t(shap2$dt[2,-1]),-apply(rDropPredDiff2,2,mean)),pch='*',
	col=c('blue','red','black'),type='b',lwd=2,lty=1,xaxt='n',ylab='SHAP',
	cex.lab=1.5,cex.axis=1.5,cex=2)
axis(1,1:length(DropVarImp),names(DropVarImp),las=2,cex.axis=1)
abline(h=0)
legend('topright',c('Kernel','Independence','Drop'),lwd=2,col=c('blue','red','black'),lty=1,cex=1.5)


# Here we'll disaggregate again, so that we can see the changes
# in values for each permutation of the feature order

par(mar=c(10,5,1,1))
matplot(-t(rDropPredDiff),type='l',col='grey',xaxt='n',ylab='')
axis(1,1:length(DropVarImp),names(DropVarImp),las=2,cex.axis=1)
lines(-apply(rDropPredDiff,2,mean),col='darkblue',lwd=5)
abline(h=0)


par(mar=c(10,5,1,1))
matplot(-t(rDropPredDiff2),type='l',col='grey',xaxt='n',ylab='')
axis(1,1:length(DropVarImp),names(DropVarImp),las=2,cex.axis=1)
lines(-apply(rDropPredDiff2,2,mean),col='darkblue',lwd=5)
abline(h=0)




####  A quick investigation of counterfactuals

MOCreg = MOCRegr$new(mod1.interp)
cfactuals = mocreg$find_counterfactuals(x_interest = train[1,],desired_outcome=log(c(400,500)))
cfactuals$data
cfactuals$subset_to_valid()
nrow(cfactuals$data)
cfactuals$data[,c(5,7:12,14)]

cfactuals$plot_freq_of_feature_changes(subset_zero = TRUE)

cfactuals$plot_parallel(feature_names = names(cfactuals$get_freq_of_feature_changes()),  digits_min_max = 2L)




## From Local to Global interpretation

# Using the IML package we can also get LIME and Shapley values 
# although the Shapley values do not have the same set of options

shaps = matrix(NA,500,18)
limes = matrix(0,500,18)
colnames(shaps) = names(test)[-1]
colnames(limes) = names(test)[-1]

for(i in 1:2){
 shaps[i,] = Shapley$new(mod1.interp,x.interest=test[i,-1])$results[,2]
 temp.lime = LocalModel$new(mod1.interp,x.interest=test[i,-1])
 limes[i,rownames(temp.lime$results)] = temp.lime$results[,1]
}

limes[is.na(limes)] = 0


# And we'll do a quick-and-dirty finite difference for saliencies

delta = apply(apply(train,2,range),2,diff)/20  # Finite differencing step is 5% of feature range

salcy = matrix(NA,500,18)
colnames(salcy) = names(ntest)[-1]

for(j in 1:18){
	ttest = test[1:500,]
	ttest[,j+1] = ttest[,j+1] + delta[j+1]
	salcy[,j] = (predict(mod1,ttest) - mod1pred[1:500])/delta[j+1]
}



# Now we can do things like look at global variable importance

par(mfrow=c(4,1))
barplot( apply( shaps^2, 2, mean) ,xaxt='n',main='Shapley Aggregation',col='blue')
barplot( apply( limes^2, 2, mean) ,xaxt='n',main='LIME Aggregation',col='blue')
barplot( apply( salcy^2, 2, mean) ,xaxt='n',main='Saliency Aggregation',col='blue')
barplot(PermImpRF,las=2,main='Permutation Importance',col='blue')


# But we note that local importance does not imply global importance. 

xx = seq(0,1,len=101)
plot(xx,xx,type='l',xlab='x1',ylab='',cex.lab=1.5,cex.axis=1.5,col='blue',lwd=2)
plot(xx,plogis(40*(xx-0.5)),type='l',xlab='x2',ylab='',cex.lab=1.5,cex.axis=1.5,col='blue',lwd=2)

dev.new()
X = runif(200)
hist( (40*plogis(40*(X-0.5))*plogis(40*(0.5-X)))^2, xlab='X',ylab='',col='blue', main='Squared Gradients')
abline(v=1,col=2,lwd=2)

median( (40*plogis(40*(X-0.5))*plogis(40*(0.5-X)))^2 )

# And we can look at relationships with other effects

par(mar=c(5,5,1,1))
plot(test[1:500,'square'],shaps[,'livingRoom'],xlab='square',ylab='livingRoom SHAP',cex.lab=1.5)
plot(test[1:500,'followers'],shaps[,'livingRoom'],xlab='followers',ylab='livingRoom SHAP',cex.lab=1.5)


plot(test[1:500,'square'],limes[,'livingRoom'],xlab='square',ylab='livingRoom LIME',cex.lab=1.5)
plot(test[1:500,'followers'],limes[,'livingRoom'],xlab='followers',ylab='livingRoom LIME',cex.lab=1.5)

plot(test[1:500,'square'],salcy[,'livingRoom'],xlab='square',ylab='livingRoom Saliency',cex.lab=1.5)
plot(test[1:500,'followers'],salcy[,'livingRoom'],xlab='followers',ylab='livingRoom Saliency',cex.lab=1.5)


# In an additive model, all of these effects should be 1:1 with their corresponding
# feature values


plot(test[1:500,'square'],shaps[,'square'],xlab='square',ylab='square SHAP',cex.lab=1.5)
plot(test[1:500,'livingRoom'],shaps[,'livingRoom'],xlab='livingRoom',ylab='livingRoom SHAP',cex.lab=1.5)
plot(test[1:500,'followers'],shaps[,'followers'],xlab='followers',ylab='followers SHAP',cex.lab=1.5)


plot(test[1:500,'square'],limes[,'square'],xlab='square',ylab='square LIME',cex.lab=1.5)
plot(test[1:500,'livingRoom'],limes[,'livingRoom'],xlab='livingRoom',ylab='livingRoom LIME',cex.lab=1.5)


plot(test[1:500,'square'],salcy[,'square'],xlab='square',ylab='square Saliency',cex.lab=1.5)
plot(test[1:500,'livingRoom'],salcy[,'livingRoom'],xlab='livingRoom',ylab='livingRoom Saliency',cex.lab=1.5)
plot(test[1:500,'followers'],salcy[,'followers'],xlab='followers',ylab='followers Saliency',cex.lab=1.5)


# And we can look at clustering say with medoids

library(cluster)
shap.cluster = pam(shaps,k=3)
lime.cluster = pam(limes,k=3)
sal.cluster = pam(limes,k=3)

# To produce a plot we will need to go back to numeric values and translate to ranks

rtest = ntest
for(j in 1:p){ 
	rtest[,j] = rank(ntest[,j])/nrow(ntest)
}


# now let's look at what we might do

par(mfrow=c(6,1),mar=c(1,1,4,1))
for(k in 1:3){
	barplot(as.numeric(rtest[shap.cluster$id.med[k],-1]),main='features',xlab='',ylab='',xaxt='n',yaxt='n',col='blue')
	barplot(shap.cluster$medoids[k,],xlab='',ylab='',xaxt='n',yaxt='n',main='Shapley Values',col='red')
}

for(k in 1:3){
	barplot(as.numeric(rtest[lime.cluster$id.med[k],-1]),main='features',xlab='',ylab='',xaxt='n',yaxt='n',col='blue')
	barplot(lime.cluster$medoids[k,],xlab='',ylab='',xaxt='n',yaxt='n',main='LIME Values',col='red')
}


for(k in 1:3){
	barplot(as.numeric(rtest[sal.cluster$id.med[k],-1]),main='features',xlab='',ylab='',xaxt='n',yaxt='n',col='blue')
	barplot(sal.cluster$medoids[k,],xlab='',ylab='',xaxt='n',yaxt='n',main='Saliency Values',col='red')
}





